# encoding: utf-8
"""
@project: sky-excel->body_parser
@author: 孙楷炎
@Email: sky4834@163.com
@synopsis: 内容键值解析器
@created_time: 2022/9/29 14:24
"""

import os
import platform

from openpyxl import load_workbook
# 头部解析器
from openpyxl.cell import MergedCell

from .parsed_cell import ParsedCell
from sky_excel.utils.j_dict import JDict


class BodyParser(object):
    wb = None
    __cell_list = []

    @property
    def cell_list(self):
        return self.__cell_list

    @cell_list.setter
    def cell_list(self, cells):
        self.__cell_list = [ParsedCell(i, self.sheet) for i in cells]

    # 获取工作工作簿实例
    def get_wb(self):
        return self.wb

    def __init__(self, file_path=None, work_book_obj=None):
        # excel对象
        if work_book_obj:
            self.wb = work_book_obj
        elif file_path and os.path.exists(file_path):
            self.wb = load_workbook(filename=file_path)
        # 共工作簿列表
        self.sheetnames = self.wb.sheetnames if self.wb else None

    def parse_body(self, sheet_name="Sheet1"):
        try:
            assert not self.wb is None, "excel 初始化失败，检出模板文件路径"
            assert sheet_name and sheet_name in self.sheetnames, "不存在名称为：%s 的工作簿" % sheet_name
            self.sheet = self.wb[sheet_name]

            # 获取合并的单元格其实单元格映射
            cells = [i for i in self.sheet.merged_cell_ranges or [] if i.max_row == self.sheet.max_row]
            # 合并单元格的开始单元格映射,版本兼容
            if platform.python_version()[0] == "3":
                start_cell_map = {i.start_cell.coordinate: i for i in cells}
            else:
                start_cell_map = {self.sheet[i.coord][0][0].coordinate: i for i in cells}

            # 最后一行代表公式行
            for cell in self.sheet[self.sheet.max_row]:
                if isinstance(cell, MergedCell) or start_cell_map.get(cell.coordinate):
                    continue
                cells.append(cell)
            # 单元格进一步封装成通用格式
            self.cell_list = cells
            return JDict({"cell_list": self.cell_list, "max_row": self.sheet.max_row}), None
        except AssertionError as logic_err:
            return None, "逻辑错误：" + str(logic_err)
        except Exception as e:
            err_detail = " file:" + str(e.__traceback__.tb_frame.f_globals["__file__"]) + " line:" + str(e.__traceback__.tb_lineno) if platform.python_version()[0] == "3" else ""
            return None, "系统异常：" + str(e) + err_detail



# _*_coding:utf-8_*_
# 推荐工具JDict搭配使用
#  folder_path = root_config.IMAGE_UPLOAD_DIR or module_config.IMAGE_UPLOAD_DIR or "/upload/image/"
import configparser
import os


class JConfig(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.__dict__ = self

    # 是否允许点操作符
    def allow_dotting(self, state=True):
        if state:
            self.__dict__ = self
        else:
            self.__dict__ = dict()

    # 当值不存在时返回None
    def __getattr__(self, *args):
        pass

    @staticmethod
    def get_section(path, section, encode="utf-8-sig"):
        if not os.path.exists(path):
            return {}

        config = configparser.ConfigParser()
        config.read(path, encoding=encode)
        if not config.has_section(section):
            return {}
        tuple_list = config.items(section)
        return {k: v for k, v in tuple_list}



# encoding: utf-8
"""
@project: sky-excel->parsed_cell
@author: 孙楷炎
@Email: sky4834@163.com
@synopsis:
@created_time: 2022/9/29 16:34
"""
from openpyxl.worksheet.merge import MergedCellRange, CellRange


class ParsedCell():
    # 表头单元格属性
    is_merge_cell = False
    boundary = None
    coordinate = None  # 坐标如：A1:B2..
    value = None

    def __init__(self, cell, sheet=None):
        if isinstance(cell, MergedCellRange):  # python3.0版本,openpyxl 3.0.* 以上版本
            self.is_merge_cell = True
            self.value = cell.start_cell.value
            self.start_cell_coord = cell.start_cell.coordinate
            self.coordinate = str(cell)
            self.row = cell.start_cell.row
            self.column = cell.start_cell.column

        elif isinstance(cell, CellRange):  # python2.0版本,openpyxl2.6.*版本
            self.is_merge_cell = True
            self.coordinate = str(cell)
            self.start_cell = sheet[self.coordinate][0][0]
            self.value = self.start_cell.value
            self.start_cell_coord = self.start_cell.coordinate
            self.row = self.start_cell.row
            self.column = self.start_cell.column

        else:  # 是普通的单元格
            self.is_merge_cell = False
            self.value = cell.value
            self.start_cell_coord = str(cell.coordinate)
            self.coordinate = str(cell.coordinate)
            self.row = cell.row
            self.column = cell.column

    def get_value(self):
        return self.value

    def set_value(self, value):
        self.value = value

    # 单元格是否在选取内判断
    # def is_in_boundary(self, cell):
    #     if not isinstance(cell, Cell):
    #         return None, "cell必须是一个单元实例"
    #     # 表头单元格实例
    #     from shapely import geometry
    #     Points = (cell.column, cell.row)
    #     top, right, bottom, left = self.boundary
    #     boundary = top + right + bottom + left
    #     line = geometry.LineString(boundary)
    #     point = geometry.Point(Points)
    #     polygon = geometry.Polygon(line)
    #     return polygon.contains(point), None

    def __str__(self):
        return self.value



# encoding: utf-8
"""
@project: sky-excel->main
@author: 孙楷炎
@Email: sky4834@163.com
@synopsis:
@created_time: 2022/9/28 18:18
"""

from sky_excel.main.v2.excel_exporter import ExcelExport

global_export_data = [
    {"col_1": "row1", "col_2": "row1", "col_3": "row1", "col_4": "row1"},
    {"col_1": "row2", "col_2": "row1", "col_3": "row1", "col_4": "row1"},
    {"col_1": "row3", "col_2": "row1", "col_3": "row1", "col_4": "row1"},
    {"col_1": "row4", "col_2": "row1", "col_3": "row1", "col_4": "row1"},
    {"col_1": "row5", "col_2": "row1", "col_3": "row1", "col_4": "row1"},
    {"col_1": "row6", "col_2": "row1", "col_3": "row1", "col_4": "row1"},
    {"col_1": "row7", "col_2": "row1", "col_3": "row1", "col_4": "row1"},
]
#
# 导出数据测试
export_instance = ExcelExport(excel_templet_path="./templet/templet.xls")
data, err = export_instance.additional_write(input_dict=global_export_data)

data, err = export_instance.save(workbook=export_instance.additional_write_wb, save_path="D:\\PySet/sky-excel/sky_excel/export/")




# encoding: utf-8
"""
@project: sky-excel->excel_exporter
@author: 孙楷炎
@Email: sky4834@163.com
@synopsis: excel输出者
@created_time: 2022/9/29 17:01
"""
from io import BytesIO
import os
import platform
import re
from string import digits
import uuid

from openpyxl import load_workbook

from .body_parser import BodyParser


class ExcelExport:
    input_dict = None
    header_cells = None
    body_map = None
    __err_msg = None
    wb = None

    @property
    def err_msg(self):
        return self.__err_msg

    @err_msg.setter
    def err_msg(self, err):
        # 返回最先暴出来的错误，按照顺序返回，避免覆盖错误无法排查
        if self.__err_msg is None:
            self.__err_msg = err

    def is_valid(self):
        return self.__err_msg

    def __init__(self, input_dict, excel_templet_path=None, excel_templet_title="Sheet1", save_path=None, ):
        # 直接解析模板，获取可操作的excel对象
        self.wb = load_workbook(filename=excel_templet_path, read_only=True) if excel_templet_path and os.path.exists(excel_templet_path) else None
        # body解析
        self.parsed_body_data, err = BodyParser(work_book_obj=self.wb).parse_body(sheet_name=excel_templet_title)
        if err:
            self.err_msg = err

        self.sheetnames = self.wb.sheetnames if self.wb else []  # 共工作簿列表
        self.sheetname = excel_templet_title
        self.save_path = save_path  # 当save_path 为空的时候则进行返回文件流的格式
        self.input_dict = input_dict
        # 合法性判断
        if not isinstance(input_dict, list):
            self.err_msg = "input_dict格式应该是[{..},{..}..]"
        if not excel_templet_path or not os.path.exists(excel_templet_path):
            self.err_msg = "excel 模板路径有误"
        if not self.sheetnames or not excel_templet_title in self.sheetnames:
            self.err_msg = "不存在名称为：%s 的工作簿" % excel_templet_title

    def write_body(self, sheet):
        try:
            # 可单独调用，不需要表头
            start_row = self.parsed_body_data.max_row or 1
            cell_list = self.parsed_body_data.cell_list or []  # 单元个对象
            # 写入数据 到数据体里面
            for row_num, row_input_dict in zip(range(start_row, len(self.input_dict) + start_row), self.input_dict):  # 行遍历
                for cell in cell_list:  # 单元格 列遍历
                    # # 数据处理、替换变量表达式
                    this_cell_value = cell.value
                    this_cell_value_match = re.compile("{{.*?}}").findall(str(cell.value) if cell.value else "")
                    this_cell_value_replaces = {i: row_input_dict.get(i.replace("{{", "").replace("}}", ""), "") for i in this_cell_value_match}
                    for k, v in this_cell_value_replaces.items():
                        this_cell_value = this_cell_value.replace(k, str(v))
                    # 数据写入
                    sheet.cell(row=row_num, column=cell.column, value=this_cell_value)
                    if cell.is_merge_cell:
                        this_cell_coordinate = [i + str(row_num) for i in cell.coordinate.translate(None, digits).split(":")]
                        sheet.merge_cells(this_cell_coordinate[0] + ":" + this_cell_coordinate[1])

            return sheet, None
        except Exception as e:
            err_detail = " file:" + str(e.__traceback__.tb_frame.f_globals["__file__"]) + " line:" + str(e.__traceback__.tb_lineno) if platform.python_version()[0] == "3" else ""
            return sheet, "写入body异常：" + str(e) + err_detail

    # 整个流程的调度函数
    def export(self):
        if self.err_msg:
            return None, self.err_msg
        try:
            # 写入数据
            writed_body_sheet, body_err = self.write_body(self.wb[self.sheetname])
            # 数据写入异常
            if body_err:
                self.err_msg = body_err
                return None, body_err
            # 保存excel
            if self.save_path:
                try:
                    # 递归创建目录
                    if not os.path.exists(self.save_path):
                        os.makedirs(self.save_path)
                    # 保存文件
                    file_path_name = (self.save_path + "/" + str(uuid.uuid3(uuid.NAMESPACE_DNS, str(uuid.uuid1()))) + ".xlsx").replace("//", "/")
                    self.wb.save(file_path_name)
                    return file_path_name, None
                except Exception as e:
                    return None, "excel 写入异常：" + str(e)
            else:  # 没有路径则返回文件流的形式
                return self.export_stream(self.wb), None
        except Exception as e:
            return None, "系统异常:" + str(e)

    # 返回文件流
    def export_stream(self, new_wb):
        try:
            from StringIO import StringIO
        except ImportError:
            from io import StringIO

        if platform.python_version()[0] == "2":
            # python 2.*.*
            output = StringIO()
            new_wb.save(output)
            output.seek(0)
            return output.getvalue()
        else:
            # python 3.*.*
            output = BytesIO()
            new_wb.save(output)
            output.seek(0)
            return output.getvalue()



# encoding: utf-8
"""
@project: sky-excel->init
@author: 孙楷炎
@Email: sky4834@163.com
@synopsis: 初始化方法
@created_time: 2022/9/28 17:22
"""

# from .main.v1.body_parser import BodyParser
# from .main.v1.excel_exporter import ExcelExport
# from .main.v1.parsed_cell import ParsedCell
#
# __all__ = ["ExcelExport", "ParsedCell", "BodyParser"]
#

from .main.v2.excel_exporter import ExcelExport

__all__ = ["ExcelExport"]



# encoding: utf-8
"""
@project: djangoModel->tool
@author: 孙楷炎
@Email: sky4834@163.com
@synopsis: CURD 工具
@created_time: 2022/6/15 14:14
"""

# ===========  工具方法 start =============
import json


# json 结果集返回
def parse_json(result):
    if not result is None:
        if type(result) is str:
            try:
                result = json.loads(result.replace("'", '"').replace('\\r', "").replace('\\n', "").replace('\\t', "").replace('\\t', ""))
            except Exception as e:
                return result
        if type(result) is list:
            for index, value in enumerate(result):
                result[index] = parse_json(value)
        if type(result) is dict:
            for k, v in result.items():
                result[k] = parse_json(v)
    return result


# 字段筛选并替换成别名
def format_params_handle(param_dict, filter_filed_list=None, remove_filed_list=None, alias_dict=None):
    # 转换的数据类型不符合，直接返回出去
    if not isinstance(param_dict, dict):
        raise Exception("param_dict 必须是字典格式")

    # 类型判断 过滤字段
    if filter_filed_list and isinstance(filter_filed_list, list):
        param_dict = {k: v for k, v in param_dict.copy().items() if k in filter_filed_list and v}

    # 类型判断， 剔除字段
    if remove_filed_list and isinstance(remove_filed_list, list):
        param_dict = {k: v for k, v in param_dict.copy().items() if not k in remove_filed_list and v}

    # 类型判断 字段转换
    if alias_dict and isinstance(alias_dict, dict):
        param_dict = {alias_dict.get(k, k): v for k, v in param_dict.copy().items()}

    return param_dict
